#pragma once
#include <iostream>
#include <memory>
#include <string>
using namespace std;
#define _DEBUG

enum class PrimaryExp_type {Exp, Number};
enum class UnaryExp_type {PrimaryExp, UnaryOp_and_UnaryExp};

static int temp_id = 0;
static string value_r = string();
inline string find_ret(const string& str){
  size_t eqPos = str.find_last_of('=');
  if (eqPos == std::string::npos) return string();
  size_t percentPos = str.rfind('%', eqPos - 1);
  size_t blankPos = str.find(' ', percentPos);
  return str.substr(percentPos,blankPos - percentPos);
}

class BaseAST {
 public:
  virtual ~BaseAST() = default;
  virtual void Dump(string& str) const = 0;
};

class CompUnitAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> func_def;

  void Dump(string& str) const override {
    #ifdef DEBUG 
    cout << "CompUnit" << endl;
    #endif
    return func_def->Dump(str);
  }
};

class FuncDefAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> func_type;
  std::string ident;
  std::unique_ptr<BaseAST> block;

  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "FuncDef" << endl;
    #endif
    str.append("fun @");
    str.append(ident);
    str.append("(): ");
    func_type->Dump(str);
    str.append(" {\n");
    block->Dump(str);
    str.append("\n}");
  }
};

class FuncTypeAST : public BaseAST {
 public:
  std::string functype = "int";

  void Dump(string& str) const override { 
    #ifdef DEBUG
    cout << "FuncType" << endl;
    #endif
    str.append("i32");
  }
};

class BlockAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> stmt;

  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "Block" << endl;
    #endif
    str.append("\%entry:\n");
    stmt->Dump(str);
  }
};

class StmtAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> exp;

  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "Stmt" << endl;
    #endif
    exp->Dump(str);
    str.append("ret "+value_r);
    value_r = string();
  }
};

class ExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> lorexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "Exp" << endl;
    #endif
    lorexp->Dump(str);
  }
};

class PrimaryExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> exp;
  std::unique_ptr<BaseAST> number;
  PrimaryExp_type type;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "PrimaryExp" << endl;
    #endif
    switch (type) {
      case PrimaryExp_type::Exp:
        exp->Dump(str);
        break;
      case PrimaryExp_type::Number:
        string nstr = "";
        number->Dump(nstr);
        value_r = nstr;
        break;
    }
  }
};

class NumberAST : public BaseAST {
 public:
  int number;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "Number" << endl;
    #endif
    str.append(to_string(number));
  }
};

class UnaryExpAST : public BaseAST {
  // 需要执行运算符，先分别取符号和运算量，然后构造表达式
  // 同时unaryexp也是被运算的量，所以任何情况都要确保有unaryexp_ret
 public:
  std::unique_ptr<BaseAST> primaryexp;
  std::unique_ptr<BaseAST> unaryop;
  std::unique_ptr<BaseAST> unaryexp;
  UnaryExp_type type;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "UnaryExp" << endl;
    #endif
    switch(type){
      case UnaryExp_type::PrimaryExp:
        primaryexp->Dump(str);
        break;
      case UnaryExp_type::UnaryOp_and_UnaryExp:

        string op = "";
        unaryop->Dump(op);
        string uexp = "";
        unaryexp->Dump(uexp);
        string uret = value_r;
        
        str.append(uexp);
        if (op=="-"){
          str.append("  %"+std::to_string(temp_id++)+" = sub 0, "+uret+"\n");
          value_r = string("%"+std::to_string(temp_id-1));
        }else if(op=="!"){
          str.append("  %"+std::to_string(temp_id++)+" = eq 0, "+uret+"\n");
          value_r = string("%"+std::to_string(temp_id-1));
        }else{
          value_r = string("%"+std::to_string(temp_id));
        }
    }
  }
};

class UnaryOpAST : public BaseAST {
 public:
  string op;
  void Dump(string& str) const override {
    str.append(op);
  }
};

class R_MulExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> r_mulexp;
  std::unique_ptr<BaseAST> unaryexp;
  string op;

  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "R_MulExp" << endl;
    #endif
    if(r_mulexp!=nullptr){
      r_mulexp->Dump(str);
      string mulr = value_r;
      R_MulExpAST* ptr = dynamic_cast<R_MulExpAST*>(r_mulexp.get());
      string r_op = ptr->op;
      unaryexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+r_op+" "+value_r+", "+mulr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      unaryexp->Dump(str);
    }
  }
};

class MulExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> r_mulexp;
  std::unique_ptr<BaseAST> unaryexp;

  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "MulExp" << endl;
    #endif
    if(r_mulexp!=nullptr)
    {
      r_mulexp->Dump(str);
      string mulr = value_r;
      R_MulExpAST* ptr = dynamic_cast<R_MulExpAST*>(r_mulexp.get());
      string r_op = ptr->op;
      unaryexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+r_op+" "+value_r+", "+mulr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      unaryexp->Dump(str);
    }
  }
};

class R_AddExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> r_addexp;
  std::unique_ptr<BaseAST> mulexp;
  string op;

  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "R_AddExp" << endl;
    #endif
    if(r_addexp!=nullptr){
      r_addexp->Dump(str);
      string addr = value_r;
      R_AddExpAST* ptr = dynamic_cast<R_AddExpAST*>(r_addexp.get());
      string r_op = ptr->op;
      mulexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+r_op+" "+value_r+", "+addr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      mulexp->Dump(str);
    }
  }
};

class AddExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> r_addexp;
  std::unique_ptr<BaseAST> mulexp;

  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "AddExp" << endl;
    #endif
    if(r_addexp!=nullptr)
    {
      r_addexp->Dump(str);
      string addr = value_r;
      R_AddExpAST* ptr = dynamic_cast<R_AddExpAST*>(r_addexp.get());
      string r_op = ptr->op;
      mulexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+r_op+" "+value_r+", "+addr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));    
    }else{
      mulexp->Dump(str);
    }
  }
 };

class R_RelExpAST : public BaseAST {
 public:
  string op;
  std::unique_ptr<BaseAST> addexp;
  std::unique_ptr<BaseAST> r_relexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "R_RelExp" << endl;
    #endif
    if(r_relexp!=nullptr)
    {
      r_relexp->Dump(str);
      string relr = value_r;
      R_RelExpAST* ptr = dynamic_cast<R_RelExpAST*>(r_relexp.get());
      string r_op = ptr->op;
      string op="";
      if(r_op=="<"){op="lt";}
      else if(r_op==">"){op="gt";}
      else if(r_op=="<="){op="le";}
      else if(r_op==">="){op="ge";}
      addexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+op+" "+value_r+", "+relr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      addexp->Dump(str);
    }
  }
};

class RelExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> addexp;
  std::unique_ptr<BaseAST> r_relexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "RelExp" << endl;
    #endif
    if(r_relexp!=nullptr)
    {
      r_relexp->Dump(str);
      string relr = value_r;
      R_RelExpAST* ptr = dynamic_cast<R_RelExpAST*>(r_relexp.get());
      string r_op = ptr->op;
      string op="";
      if(r_op=="<"){op="lt";}
      else if(r_op==">"){op="gt";}
      else if(r_op=="<="){op="le";}
      else if(r_op==">="){op="ge";}
      addexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+op+" "+value_r+", "+relr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      addexp->Dump(str);
    }
  }
};

class R_EqExpAST : public BaseAST {
 public:
  string op;
  std::unique_ptr<BaseAST> relexp;
  std::unique_ptr<BaseAST> r_eqexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "R_EqExp" << endl;
    #endif
    if(r_eqexp!=nullptr)
    {
      r_eqexp->Dump(str);
      string eqr = value_r;
      R_EqExpAST* ptr = dynamic_cast<R_EqExpAST*>(r_eqexp.get());
      string r_op = ptr->op;
      string op="";
      if(r_op=="=="){op="eq";}
      else if(r_op=="!="){op="ne";}
      relexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+op+" "+value_r+", "+eqr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      relexp->Dump(str);
    }
  }
};

class EqExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> relexp;
  std::unique_ptr<BaseAST> r_eqexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "EqExp" << endl;
    #endif
    if(r_eqexp!=nullptr)
    {
      r_eqexp->Dump(str);
      string eqr = value_r;
      R_EqExpAST* ptr = dynamic_cast<R_EqExpAST*>(r_eqexp.get());
      string r_op = ptr->op;
      string op="";
      if(r_op=="=="){op="eq";}
      else if(r_op=="!="){op="ne";}
      relexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+op+" "+value_r+", "+eqr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      relexp->Dump(str);
    }
  }
};

class R_LAndExpAST : public BaseAST {
 public:
  string op;
  std::unique_ptr<BaseAST> eqexp;
  std::unique_ptr<BaseAST> r_landexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "R_LAndExp" << endl;
    #endif
    if(r_landexp!=nullptr)
    {
      r_landexp->Dump(str);
      string landr = value_r;
      string op="&&";
      eqexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+op+" "+value_r+", "+landr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      eqexp->Dump(str);
    }
  }
};


class LAndExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> eqexp;
  std::unique_ptr<BaseAST> r_landexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "LAndExp" << endl;
    #endif
    if(r_landexp!=nullptr)
    {
      r_landexp->Dump(str);
      string landr = value_r;
      string op="&&";
      eqexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+op+" "+value_r+", "+landr+"\n");
      value_r = string("%"+std::to_string(temp_id-1));
    }else{
      eqexp->Dump(str);
    }
  }
};

class R_LOrExpAST : public BaseAST {
 public:
  string op;
  std::unique_ptr<BaseAST> landexp;
  std::unique_ptr<BaseAST> r_lorexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "R_LOrExp" << endl;
    #endif
    if(r_lorexp!=nullptr)
    {
      r_lorexp->Dump(str);
      string lorr = value_r;
      string op="||";
      landexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+op+" "+value_r+", "+lorr+"\n");
      value_r =string("%"+std::to_string(temp_id-1));
    }else{
      landexp->Dump(str);
    }
  }
};

class LOrExpAST : public BaseAST {
 public:
  std::unique_ptr<BaseAST> landexp;
  std::unique_ptr<BaseAST> r_lorexp;
  void Dump(string& str) const override {
    #ifdef DEBUG
    cout << "LOrExp" << endl;
    #endif
    if(r_lorexp!=nullptr)
    {
      r_lorexp->Dump(str);
      string lorr = value_r;
      string op="||";
      landexp->Dump(str);
      str.append("  %"+std::to_string(temp_id++)+" = "+op+" "+value_r+", "+lorr+"\n");
      value_r =string("%"+std::to_string(temp_id-1));
    }else{
      landexp->Dump(str);
    }
  }
};

