%code requires {
  #include <memory>
  #include <string>
  #include "../ast_class.h"
}

%{

#include <iostream>
#include <memory>
#include <string>
#include "../ast_class.h"

int yylex();
void yyerror(std::unique_ptr<BaseAST> &ast, const char *s);

using namespace std;

%}


%parse-param { std::unique_ptr<BaseAST> &ast }


%union {
  std::string *str_val;
  int int_val;
  BaseAST* ast_val;
}

%token INT RETURN
%token <str_val> IDENT RELOP EQOP LANDOP LOROP
%token <int_val> INT_CONST

%type <ast_val> FuncDef FuncType Block Stmt Exp PrimaryExp UnaryExp UnaryOp    
%type <ast_val> Number MulExp AddExp R_MulExp R_AddExp RelExp R_RelExp EqExp R_EqExp LAndExp R_LAndExp LOrExp R_LOrExp
%%

CompUnit
  : FuncDef {
    auto comp_unit = make_unique<CompUnitAST>();
    comp_unit->func_def = unique_ptr<BaseAST>($1);
    ast = move(comp_unit);
  }
  ;

FuncDef
  : FuncType IDENT '(' ')' Block {
    auto ast = new FuncDefAST();
    ast->func_type = unique_ptr<BaseAST>($1);
    ast->ident = *unique_ptr<string>($2);
    ast->block = unique_ptr<BaseAST>($5);
    $$ = ast;
  }
  ;

FuncType
  : INT {
    auto ast = new FuncTypeAST();
    $$ = ast;
  }
  ;

Block
  : '{' Stmt '}' {
    auto ast = new BlockAST();
    ast->stmt = unique_ptr<BaseAST>($2);
    $$ = ast;
  }
  ;

Stmt
  : RETURN Exp ';' {
    auto ast = new StmtAST();
    ast->exp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }
  ;

Exp 
  : LOrExp {
    auto ast = new ExpAST();
    ast->lorexp = unique_ptr<BaseAST>($1);
    $$ = ast;
  }
  ;

PrimaryExp
  : '(' Exp ')' {
    auto ast = new PrimaryExpAST();
    ast->type = PrimaryExp_type::Exp;
    ast->exp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }
  | Number {
    auto ast = new PrimaryExpAST();
    ast->type = PrimaryExp_type::Number;
    ast->number = unique_ptr<BaseAST>($1);
    $$ = ast;
  }
  ;

Number
  : INT_CONST {
    auto ast = new NumberAST();
    ast->number = $1;
    $$ = ast;
  }
  ;

UnaryExp
  : PrimaryExp {
    auto ast = new UnaryExpAST();
    ast->type = UnaryExp_type::PrimaryExp;
    ast->primaryexp = unique_ptr<BaseAST>($1);
    $$ = ast;
  }
  | UnaryOp UnaryExp {
    auto ast = new UnaryExpAST();
    ast->type = UnaryExp_type::UnaryOp_and_UnaryExp;
    ast->unaryop = unique_ptr<BaseAST>($1);
    ast->unaryexp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }
  ;

UnaryOp
  :'+' {
     auto ast = new UnaryOpAST();
     ast->op = "+";
     $$ = ast;
  } 
  |'-' {
     auto ast = new UnaryOpAST();
     ast->op = "-";
     $$ = ast;
  }
  |'!' {
     auto ast = new UnaryOpAST();
     ast->op = "!";
     $$ = ast;
  }

MulExp
  : UnaryExp R_MulExp {
    auto ast = new MulExpAST();
    ast->unaryexp = unique_ptr<BaseAST>($1);
    ast->r_mulexp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }

R_MulExp
  : {$$ = nullptr;}
  | '*' UnaryExp R_MulExp {
    auto ast = new R_MulExpAST();
    ast->op = "mul";
    ast->unaryexp = unique_ptr<BaseAST>($2);
    ast->r_mulexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }
  | '/' UnaryExp R_MulExp {
    auto ast = new R_MulExpAST();
    ast->op = "div";
    ast->unaryexp = unique_ptr<BaseAST>($2);
    ast->r_mulexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }
  | '%' UnaryExp R_MulExp {
    auto ast = new R_MulExpAST();
    ast->op = "mod";
    ast->unaryexp = unique_ptr<BaseAST>($2);
    ast->r_mulexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }

AddExp
  : MulExp R_AddExp {
    auto ast = new AddExpAST();
    ast->mulexp = unique_ptr<BaseAST>($1);
    ast->r_addexp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }

R_AddExp
  : {$$ = nullptr;}
  | '+' MulExp R_AddExp{
    auto ast = new R_AddExpAST();
    ast->op = "add";
    ast->mulexp = unique_ptr<BaseAST>($2);
    ast->r_addexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }
  | '-' MulExp R_AddExp {
    auto ast = new R_AddExpAST();
    ast->op = "sub";
    ast->mulexp = unique_ptr<BaseAST>($2);
    ast->r_addexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }

RelExp
  : AddExp R_RelExp {
    auto ast = new RelExpAST();
    ast->addexp = unique_ptr<BaseAST>($1);
    ast->r_relexp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }

R_RelExp
  : {$$ = nullptr;}
  | RELOP AddExp R_RelExp {
    auto ast = new R_RelExpAST();
    ast->op = *unique_ptr<string>($1);
    ast->addexp = unique_ptr<BaseAST>($2);
    ast->r_relexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }

EqExp
  : RelExp R_EqExp {
    auto ast = new EqExpAST();
    ast->relexp = unique_ptr<BaseAST>($1);
    ast->r_eqexp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }

R_EqExp
  : {$$ = nullptr;}
  | EQOP RelExp R_EqExp {
    auto ast = new R_EqExpAST();
    ast->op = *unique_ptr<string>($1);
    ast->relexp = unique_ptr<BaseAST>($2);
    ast->r_eqexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }

LAndExp
  : EqExp R_LAndExp {
    auto ast = new LAndExpAST();
    ast->eqexp = unique_ptr<BaseAST>($1);
    ast->r_landexp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }

R_LAndExp
  : {$$ = nullptr;}
  | LANDOP EqExp R_LAndExp {
    auto ast = new R_LAndExpAST();
    ast->op =  *unique_ptr<string>($1);
    ast->eqexp = unique_ptr<BaseAST>($2);
    ast->r_landexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }

LOrExp
  : LAndExp R_LOrExp {
    auto ast = new LOrExpAST();
    ast->landexp = unique_ptr<BaseAST>($1);
    ast->r_lorexp = unique_ptr<BaseAST>($2);
    $$ = ast;
  }

R_LOrExp
  : {$$ = nullptr;}
  | LOROP LAndExp R_LOrExp {
    auto ast = new R_LOrExpAST();
    ast->op = *unique_ptr<string>($1);
    ast->landexp = unique_ptr<BaseAST>($2);
    ast->r_lorexp = unique_ptr<BaseAST>($3);
    $$ = ast;
  }

%%

// parser 如果发生错误 (例如输入的程序出现了语法错误), 就会调用这个函数
void yyerror(unique_ptr<BaseAST> &ast, const char *s) {
  cerr << "error: " << s << endl;
}
